# sage_setup: distribution = sagemath-combinat
r"""
T-sequences

T-sequences are tuples of four (-1, 0, 1) sequences of length `t` where
for every `i` exactly one sequence has a nonzero entry at index `i`
and for which the nonperiodic autocorrelation function is equal to zero
(i.e. they are complementary). See Definition 7.5 of [Seb2017]_.

These can be constructed from Turyn sequences. In particular,
if Turyn sequences of length `l` exists, there will be T-sequences
of length `4l-1` and `2l-1`.

Turyn sequences are tuples of four (-1, +1) sequences `X, U, Y, V` of length
`l`, `l`, `l-1`, `l-1` with nonperiodic autocorrelation equal to zero and
the additional constraints that:

* the first element of `X` is 1
* the last element of `X` is -1
* the last element of `U` is 1

The nonperiodic autocorrelation of a family of sequences
`X=\{A_1, A_2, ..., A_n\}` is defined as (see Definition 7.2 of [Seb2017]_):

.. MATH::

    N_X(j) = \sum_{i=1}^{n-j}(a_{1,i}a_{1,i+j} + a_{2,i}a_{2,i+j} + ... + a_{n,i}a_{n,i+j})

AUTHORS:

- Matteo Cati (2022-11-16): initial version
"""

# ***************************************************************************
#       Copyright (C) 2022 Matteo Cati matteo.cati@keble.ox.ac.uk
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.structure.sequence import Sequence


def _nonperiodic_autocorrelation(sequences, j):
    r"""
    Compute the nonperiodic autocorrelation of a family of sequences.

    Namely, given a family of sequences `X` it computes:

    .. MATH::

        N_X(j) = \sum_{i=1}^{n-j}(a_{1,i}a_{1,i+j} + a_{2,i}a_{2,i+j} + ... + a_{n,i}a_{n,i+j})

    INPUT:

    - ``sequences`` -- either a single sequence or a list of sequences for
      which we want to compute the nonperiodic autocorrelation

    - ``j`` -- integer; the parameter `j` used when calculating the nonperiodic
      autocorrelation
    """
    if not isinstance(sequences[0], list):
        sequences = [sequences]

    t = len(sequences[0])
    result = 0
    for i in range(t - j):
        for seq in sequences:
            result += seq[i] * seq[i + j]
    return result


def is_skew(seq, verbose=False):
    r"""
    Check if the given sequence is skew.

    A sequence `X=\{x_1, x_2, ...,x_n\}` is defined skew (according to Definition
    7.4 of [Seb2017]_) if `n` is even and `x_i = -x_{n-i+1}`.

    INPUT:

    - ``seq`` -- the sequence that should be checked

    - ``verbose`` -- boolean (default: ``False``); if ``True`` the function
      will be verbose when the sequences do not satisfy the constraints

    EXAMPLES::

        sage: from sage.combinat.t_sequences import is_skew
        sage: is_skew([1, -1, 1, -1, 1, -1])
        True
        sage: is_skew([1, -1, -1, -1], verbose=True)
        Constraint not satisfied at index 1
        False

    TESTS::

        sage: is_skew([1, -1, -1])
        False
        sage: is_skew([1, -1, -1, 1, -1], verbose=True)
        Sequence should be of even length
        False
    """
    n = len(seq)

    if n % 2:
        if verbose:
            print('Sequence should be of even length')
        return False

    for i in range(n):
        if seq[i] != -seq[n-i-1]:
            if verbose:
                print(f'Constraint not satisfied at index {i}')
            return False
    return True


def is_symmetric(seq, verbose=False) -> bool:
    r"""
    Check if the given sequence is symmetric.

    A sequence `X=\{x_1, x_2, ...,x_n\}` is defined symmetric (according to Definition
    7.4 of [Seb2017]_) if `n` is odd and `x_i = x_{n-i+1}`.

    INPUT:

    - ``seq`` -- the sequence that should be checked

    - ``verbose`` -- boolean (default: ``False``); if ``True`` the function will be
      verbose when the sequences do not satisfy the constraints

    EXAMPLES::

        sage: from sage.combinat.t_sequences import is_symmetric
        sage: is_symmetric([1, -1, 1, -1, 1])
        True
        sage: is_symmetric([1, -1, 1, 1, 1], verbose=True)
        Constraint not satisfied at index 1
        False

    TESTS::

        sage: is_symmetric([1, -1, -1, 1])
        False
        sage: is_symmetric([1, -1, -1, 1], verbose=True)
        Sequence should be of odd length
        False
    """
    n = len(seq)

    if n % 2 == 0:
        if verbose:
            print('Sequence should be of odd length')
        return False

    for i in range(n):
        if seq[i] != seq[n-i-1]:
            if verbose:
                print(f'Constraint not satisfied at index {i}')
            return False
    return True


def is_T_sequences_set(sequences, verbose=False):
    r"""
    Check if a family of sequences is composed of T-sequences.

    Given 4 (-1, 0, +1) sequences, they will be T-sequences if
    (Definition 7.4 of [Seb2017]_):

    * they have all the same length `t`
    * for each index `i`, exactly one sequence is nonzero at `i`
    * the nonperiodic autocorrelation is equal to `0`

    INPUT:

    - ``sequences`` -- list of four sequences

    - ``verbose`` -- boolean (default: ``False``); if ``True`` the function will be
      verbose when the sequences do not satisfy the constraints

    EXAMPLES::

        sage: from sage.combinat.t_sequences import is_T_sequences_set
        sage: seqs = [[1, 1, 0, 0, 0], [0, 0, 1, 0, 0], [0, 0, 0, 1, -1], [0, 0, 0, 0, 0]]
        sage: is_T_sequences_set(seqs)
        True
        sage: seqs = [[1, 1, 0, 1, 0], [0, 0, 1, 0, 0], [0, 0, 0, 1, -1], [0, 0, 0, 0, 0]]
        sage: is_T_sequences_set(seqs, verbose=True)
        There should be exactly a nonzero element at every index, found 2 such
         elements at index 3
        False


    TESTS::

        sage: seqs = [[1, 1, 0, 1, 0], [0, 0, 1, 0, 0], [0, 0, 0, 1, -1], [0, 0, 0, 0, 0]]
        sage: is_T_sequences_set(seqs)
        False
        sage: seqs = [[1, 1, 0, 0, 0], [0, 0, 1, 0, 0], [0, 0, 0, -1, -1], [0, 0, 0, 0, 0]]
        sage: is_T_sequences_set(seqs, verbose=True)
        Nonperiodic autocorrelation should always be zero, found 2 for parameter 1
        False
        sage: is_T_sequences_set([[1, 0, ], [0, -1, 0], [0, 0, 1]])
        False
        sage: seqs = [[1, 2, 0, 0, 0], [0, 0, 1, 0, 0], [0, 0, 0, -1, -1], [0, 0, 0, 0, 0]]
        sage: is_T_sequences_set(seqs, verbose=True)
        Elements should be in (-1, 0, +1), but 2 was found at index 1
        False
    """
    if len(sequences) != 4:
        if verbose:
            print(f"T-Sequence should contain 4 sequences, found {len(sequences)} instead")
        return False

    t = len(sequences[0])

    for i in range(t):
        tot = 0
        for seq in sequences:
            if seq[i] not in [-1, 0, 1]:
                if verbose:
                    print(f"Elements should be in (-1, 0, +1), but {seq[i]} was found at index {i}")
                return False
            tot += abs(seq[i])
        if tot != 1:
            if verbose:
                print(f"There should be exactly a nonzero element at every index, found {tot} such elements at index {i}")
            return False

    for j in range(1, t):
        autocorr = _nonperiodic_autocorrelation(sequences, j)
        if autocorr != 0:
            if verbose:
                print(f"Nonperiodic autocorrelation should always be zero, found {autocorr} for parameter {j}")
            return False

    return True


def turyn_sequences_smallcases(l, existence=False):
    r"""
    Construction of Turyn sequences for small values of `l`.

    The data is taken from [Seb2017]_ and [CRSKKY1989]_.

    INPUT:

    - ``l`` -- integer; the length of the Turyn sequences

    - ``existence`` -- boolean (default: ``False``); if ``True``, only return
      whether the Turyn sequences are available for the given length

    EXAMPLES:

    By default, this method returns the four Turyn sequences ::

        sage: from sage.combinat.t_sequences import turyn_sequences_smallcases
        sage: turyn_sequences_smallcases(4)
        [[1, 1, -1, -1], [1, 1, -1, 1], [1, 1, 1], [1, -1, 1]]

    If we pass the ``existence`` flag, the method will return a boolean ::

        sage: turyn_sequences_smallcases(4, existence=True)
        True

    TESTS::

        sage: turyn_sequences_smallcases(17)
        Traceback (most recent call last):
        ...
        ValueError: Turyn sequence of length 17 is not implemented yet.
        sage: turyn_sequences_smallcases(17, existence=True)
        False
    """
    db = {
        2: [[1, -1], [1, 1], [1], [1]],
        3: [[1, 1, 1], [1, 1, -1], [1, -1], [1, -1]],
        4: [[1, 1, -1, -1], [1, 1, -1, 1], [1, 1, 1], [1, -1, 1]],
        5: [[1, 1, -1, 1, 1], [1, 1, 1, 1, -1], [1, 1, -1, -1], [1, -1, 1, -1]],
        6: [[1, 1, 1, -1, -1, -1], [1, 1, -1, 1, -1, 1], [1, 1, -1, 1, 1], [1, 1, -1, 1, 1]],
        7: [[1, 1, 1, -1, 1, 1, 1], [1, 1, -1, -1, -1, 1, -1], [1, 1, -1, 1, -1, -1], [1, 1, -1, 1, -1, -1]],
        8: [[1, 1, -1, 1, -1, 1, -1, -1], [1, 1, 1, 1, -1, -1, -1, 1], [1, 1, 1, -1, 1, 1, 1], [1, -1, -1, 1, -1, -1, 1]],
        13: [[1, 1, 1, 1, -1, 1, -1, 1, -1, 1, 1, 1, 1], [1, 1, 1, -1, -1, 1, -1, 1, -1, -1, 1, 1, -1],
             [1, 1, 1, -1, 1, 1, -1, -1, 1, -1, -1, -1], [1, 1, 1, -1, -1, 1, -1, 1, 1, -1, -1, -1]],
        15: [[1, 1, -1, 1, 1, 1, -1, 1, -1, 1, 1, 1, -1, 1, 1], [1, 1, 1, -1, 1, 1, -1, -1, -1, 1, 1, -1, 1, 1, -1],
             [1, 1, 1, 1, -1, -1, 1, -1, 1, 1, -1, -1, -1, -1], [1, -1, -1, -1, -1, 1, -1, 1, -1, 1, 1, 1, 1, -1]],
    }

    if existence:
        return l in db

    if l not in db:
        raise ValueError(f"Turyn sequence of length {l} is not implemented yet.")

    return list(map(Sequence, db[l]))


def T_sequences_construction_from_base_sequences(base_sequences, check=True):
    r"""
    Construct T-sequences of length `2n+p` from base sequences of length `n+p, n+p, n, n`.

    Given base sequences `A, B, C, D`, the T-sequences are constructed as described in
    [KTR2005]_:

    .. MATH::

        \begin{aligned}
        T_1 &= \frac{1}{2}(A+B); 0_{n} \\
        T_2 &= \frac{1}{2}(A-B); 0_{n} \\
        T_3 &= 0_{n+p} + \frac{1}{2}(C+D) \\
        T_4 &= 0_{n+p} + \frac{1}{2}(C-D)
        \end{aligned}

    INPUT:

    - ``base_sequences`` -- the base sequences that should be used to construct
      the T-sequences

    - ``check`` -- boolean (default: ``True``); check that the sequences
      created are T-sequences before returning them

    EXAMPLES::

        sage: from sage.combinat.t_sequences import turyn_sequences_smallcases, T_sequences_construction_from_base_sequences
        sage: seqs = turyn_sequences_smallcases(4)
        sage: T_sequences_construction_from_base_sequences(seqs)
        [[1, 1, -1, 0, 0, 0, 0],
        [0, 0, 0, -1, 0, 0, 0],
        [0, 0, 0, 0, 1, 0, 1],
        [0, 0, 0, 0, 0, 1, 0]]

    TESTS::

        sage: from sage.combinat.t_sequences import base_sequences_construction, is_T_sequences_set
        sage: seqs = turyn_sequences_smallcases(4)
        sage: is_T_sequences_set(T_sequences_construction_from_base_sequences(seqs))
        True
        sage: T_sequences_construction_from_base_sequences([[1, -1], [-1, 1], [1]])
        Traceback (most recent call last):
        ...
        AssertionError
        sage: X = [1,1,-1,1,-1,1,-1,1]
        sage: Y = [1,-1,-1,-1,-1,-1,-1,1]
        sage: Z = [1,-1,-1,1,1,1,1,-1]
        sage: W = [1,1,1,-1,1,1,-1]
        sage: base_seqs = base_sequences_construction([X, Y, Z, W])
        sage: is_T_sequences_set(T_sequences_construction_from_base_sequences(base_seqs))
        True
    """

    assert len(base_sequences) == 4

    A, B, C, D = base_sequences
    n = len(C)
    p = len(A)-n

    assert len(A) == len(B) == len(C)+p == len(D)+p

    def seq_sum(seq1, seq2):
        return [(a+b)//2 for (a, b) in zip(seq1, seq2)]

    def seq_subtract(seq1, seq2):
        return [(a-b)//2 for (a, b) in zip(seq1, seq2)]

    def zero_seq(n):
        return [0 for _ in range(n)]

    X1 = Sequence(seq_sum(A, B) + zero_seq(n))
    X2 = Sequence(seq_subtract(A, B) + zero_seq(n))
    X3 = Sequence(zero_seq(n+p) + seq_sum(C, D))
    X4 = Sequence(zero_seq(n+p) + seq_subtract(C, D))

    res = [X1, X2, X3, X4]
    if check:
        assert is_T_sequences_set(res)
    return res


def T_sequences_construction_from_turyn_sequences(turyn_sequences, check=True):
    r"""
    Construct T-sequences of length `4l-1` from Turyn sequences of length `l`.

    Given Turyn sequences `X, U, Y, V`, the T-sequences are constructed as described in
    theorem 7.7 of [Seb2017]_:

    .. MATH::

        \begin{aligned}
        T_1 &= 1; 0_{4l-2} \\
        T_2 &= 0; X/Y; 0_{2l-1} \\
        T_3 &= 0_{2l}; U/0_{l-2} \\
        T_4 &= 0_{2l} + 0_{l}/V
        \end{aligned}

    INPUT:

    - ``turyn_sequences`` -- the Turyn sequences that should be used to
      construct the T-sequences

    - ``check`` -- boolean (default: ``True``); check that the sequences
      created are T-sequences before returning them

    EXAMPLES::

        sage: from sage.combinat.t_sequences import turyn_sequences_smallcases, T_sequences_construction_from_turyn_sequences, is_T_sequences_set
        sage: seqs = turyn_sequences_smallcases(4)
        sage: T_sequences_construction_from_turyn_sequences(seqs)
        [[1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 1, 1, 1, 1, -1, 1, -1, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, -1, 0, 1],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, -1, 0, 1, 0]]

    TESTS::

        sage: seqs = turyn_sequences_smallcases(4)
        sage: is_T_sequences_set(T_sequences_construction_from_turyn_sequences(seqs))
        True
        sage: T_sequences_construction_from_turyn_sequences([[1, -1], [-1, 1], [1]])
        Traceback (most recent call last):
        ...
        AssertionError
    """

    assert len(turyn_sequences) == 4

    X, U, Y, V = turyn_sequences
    l = len(X)

    assert len(X) == len(U) == len(Y)+1 == len(V)+1

    def zero_seq(n):
        return [0 for _ in range(n)]

    def interleave(seq1, seq2):
        res = []
        for i in range(len(seq1) + len(seq2)):
            if i % 2 == 0:
                res.append(seq1[i//2])
            else:
                res.append(seq2[i//2])
        return res

    X1 = Sequence([1] + zero_seq(4*l-2))
    X2 = Sequence([0] + interleave(X, Y) + zero_seq(2*l-1))
    X3 = Sequence(zero_seq(2*l) + interleave(U, zero_seq(l-1)))
    X4 = Sequence(zero_seq(2*l) + interleave(zero_seq(l), V))

    res = [X1, X2, X3, X4]
    if check:
        assert is_T_sequences_set(res)
    return res


def T_sequences_smallcases(t, existence=False, check=True):
    r"""
    Construct T-sequences for some small values of `t`.

    This method will try to use the constructions defined in
    :func:`T_sequences_construction_from_base_sequences` and
    :func:`T_sequences_construction_from_turyn_sequences`
    together with the Turyn sequences stored in :func:`turyn_sequences_smallcases`,
    or base sequences created by :func:`base_sequences_smallcases`.

    This function contains also some T-sequences taken directly from [CRSKKY1989]_.

    INPUT:

    - ``t`` -- integer; the length of the T-sequences to construct

    - ``existence`` -- boolean (default: ``False``); if ``True``, this method
      only returns whether a T-sequences of the given size can be constructed

    - ``check`` -- boolean (default: ``True``); check that the sequences are
      T-sequences before returning them

    EXAMPLES:

    By default, this method returns the four T-sequences ::

        sage: from sage.combinat.t_sequences import T_sequences_smallcases, is_T_sequences_set
        sage: T_sequences_smallcases(9)
        [[1, 1, 0, 1, 0, 0, 0, 0, 0],
        [0, 0, -1, 0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 1, 0, 0, -1],
        [0, 0, 0, 0, 0, 0, 1, -1, 0]]

    If the existence flag is passed, the method returns a boolean ::

        sage: T_sequences_smallcases(9, existence=True)
        True

    TESTS::

        sage: T_sequences_smallcases(66)
        Traceback (most recent call last):
        ...
        ValueError: T Sequences of length 66 not yet implemented.
        sage: is_T_sequences_set(T_sequences_smallcases(47))
        True
        sage: is_T_sequences_set(T_sequences_smallcases(11))
        True
        sage: T_sequences_smallcases(69, existence=True)
        False
    """
    db = {
        47: [
            [1,-1,-1,0,0,-1,1,-1]+[0]*8+[1,-1,-1,0,0,-1,-1]+[0]*24,
            [0,0,0,-1,1,0,0,0,-1,-1,-1,1,1,1,1,1,0,0,0,1,-1,0,0,1]+[0]*23,
            [0]*26+[-1,0,1,0,0,0,0,1,-1,1,1,1,0,0,0,0,1,0,-1,0,0],
            [0]*24 + [1,1,0,-1,0,-1,1,1,-1,0,0,0,0,0,-1,1,-1,-1,0,-1,0,-1,1]
        ],
        65: [
            [0]*33+[1,1,1,1,1,-1,-1,1,1,-1,1,-1,1,1,-1,-1,1,1,1,1,1,-1,-1,1,-1,1,-1,1,-1,-1,1,1],
            [0]*32+[1]+[0]*32,
            [1]*5+[-1,-1,1,1,-1,1,-1,1,1]+[-1]*7+[1,1,-1,1,-1,1,-1,1,1,-1,-1]+[0]*33,
            [0]*65
        ],
        93: [
            [0,-1,0,0,-1,1,0,-1,1,0,1,1,0,0,1,1,1,0,0,-1,0,-1,1,1,1,-1,0,1,0,0,1]+[0]*33+[1,1,0,0,1,0,0,-1,0,0,-1,1,0,1]+[0]*15,
            [-1,0,-1,1,0,0,1,0,0,-1,0,0,-1,-1,0,0,0,-1,1,0,1]+[0]*5+[-1,0,1,1]+[0]*32+[1,1,0,0,1,1,0,1,-1,0,1,-1,0,0,-1]+[0]*16,
            [0]*32+[1,0,0,1,-1,0,1,-1,0,-1,-1,0,0,-1,-1,1,0,0,-1,0,-1,1,1,1,-1,0,1,0,0,1]+[0]*17+[1,1,0,-1]+[0]*5+[1,0,1,-1,0],
            [0]*31+[1,0,1,-1,0,0,-1,0,0,1,0,0,1,1,0,0,0,-1,1,0,1]+[0]*5+[-1,0,1,1]+[0]*17+[-1,0,0,-1,0,1,-1,-1,-1,1,0,1,0,0,-1]
        ]
    }

    if t in db:
        if existence:
            return True
        sequences = list(map(Sequence, db[t]))
        if check:
            assert is_T_sequences_set(sequences)
        return sequences
    if (t+1) % 2 == 0 and turyn_sequences_smallcases((t+1)//2, existence=True):
        if existence:
            return True
        turyn_seqs = turyn_sequences_smallcases((t+1)//2)
        return T_sequences_construction_from_base_sequences(turyn_seqs, check=check)

    if (t+1) % 4 == 0 and turyn_sequences_smallcases((t+1)//4, existence=True):
        if existence:
            return True
        turyn_seqs = turyn_sequences_smallcases((t+1)//4)
        return T_sequences_construction_from_turyn_sequences(turyn_seqs, check=check)

    for p in range(1, t):
        n = (t-p)//2
        if (t-p) % 2 == 0 and base_sequences_smallcases(n, p, existence=True):
            if existence:
                return True
            base_seqs = base_sequences_smallcases(n, p, check=False)
            return T_sequences_construction_from_base_sequences(base_seqs, check=check)

    if existence:
        return False
    raise ValueError(f'T Sequences of length {t} not yet implemented.')


def base_sequences_construction(turyn_type_seqs, check=True):
    r"""
    Construct base sequences of length `2n-1, 2n-1, n, n` from Turyn type
    sequences of length `n,n,n,n-1`.

    Given Turyn type sequences `X, Y, Z, W` of length `n,n,n,n-1`, Theorem 1 of
    [KTR2005]_  shows that the following are base sequences of length
    `2n-1, 2n-1, n, n`:

    .. MATH::

        \begin{aligned}
        A &= Z;W \\
        B &= Z; -W \\
        C &= X \\
        D &= Y
        \end{aligned}

    INPUT:

    - ``turyn_type_seqs`` -- the list of 4 Turyn type sequences that should be
      used to construct the base sequences

    - ``check`` -- boolean (default: ``True``); check that the resulting
      sequences are base sequences before returning them

    OUTPUT: list containing the four base sequences

    EXAMPLES::

        sage: from sage.combinat.t_sequences import base_sequences_construction
        sage: X = [1,1,-1,1,-1,1,-1,1]
        sage: Y = [1,-1,-1,-1,-1,-1,-1,1]
        sage: Z = [1,-1,-1,1,1,1,1,-1]
        sage: W = [1,1,1,-1,1,1,-1]
        sage: base_sequences_construction([X, Y, Z, W])
        [[1, -1, -1, 1, 1, 1, 1, -1, 1, 1, 1, -1, 1, 1, -1],
        [1, -1, -1, 1, 1, 1, 1, -1, -1, -1, -1, 1, -1, -1, 1],
        [1, 1, -1, 1, -1, 1, -1, 1],
        [1, -1, -1, -1, -1, -1, -1, 1]]

    TESTS::

        sage: base_sequences_construction([[1, -1], [1], [1], [-1]])
        Traceback (most recent call last):
        ...
        AssertionError

    .. SEEALSO::

        :func:`is_base_sequences_tuple`
    """
    assert len(turyn_type_seqs) == 4
    X, Y, Z, W = turyn_type_seqs

    assert len(X) == len(Y) == len(Z) == len(W)+1

    A = Sequence(Z + W)
    B = Sequence(Z + [-el for el in W])
    C = X
    D = Y

    if check:
        assert is_base_sequences_tuple([A, B, C, D])
    return [A, B, C, D]


def is_base_sequences_tuple(base_sequences, verbose=False):
    r"""Check if the given sequences are base sequences.

    Four (-1, +1) sequences `A, B, C, D` of length `n+p, n+p, n, n` are called
    base sequences if for all `j \ge 1`:

    .. MATH::

        N_A(j)+N_B(j)+N_C(j)+N_D(j) = 0

    where `N_X(j)` is the nonperiodic autocorrelation (See definition in [KTR2005]_).

    INPUT:

    - ``base_sequences`` -- the list of 4 sequences that should be checked

    - ``verbose`` -- boolean (default: ``False``); if ``True`` the function
      will be verbose when the sequences do not satisfy the constraints

    EXAMPLES::

        sage: from sage.combinat.t_sequences import is_base_sequences_tuple
        sage: seqs = [[1, -1, -1, 1, 1, 1, 1, -1, 1, 1, 1, -1, 1, 1, -1],[1, -1, -1, 1, 1, 1, 1, -1, -1, -1, -1, 1, -1, -1, 1],[1, 1, -1, 1, -1, 1, -1, 1],[1, -1, -1, -1, -1, -1, -1, 1]]
        sage: is_base_sequences_tuple(seqs)
        True

    If verbose is true, the function will be verbose ::

        sage: seqs = [[1, -1], [1, 1], [-1], [2]]
        sage: is_base_sequences_tuple(seqs, verbose=True)
        Base sequences should only contain -1, +1, found 2
        False

    TESTS:

        sage: seqs = [[1, -1], [1], [-1]]
        sage: is_base_sequences_tuple(seqs)
        False
        sage: seqs = [[1, -1], [1, -1], [-1], [1]]
        sage: is_base_sequences_tuple(seqs)
        False
        sage: seqs = [[1, -1], [1, 1], [-1], [2]]
        sage: is_base_sequences_tuple(seqs)
        False
        sage: seqs = [[1, -1], [1], [-1], [1]]
        sage: is_base_sequences_tuple(seqs)
        False

    .. SEEALSO::

        :func:`base_sequences_construction`
    """
    if len(base_sequences) != 4:
        if verbose:
            print(f'Base sequences should be 4, found {len(base_sequences)}')
        return False
    A, B, C, D = base_sequences
    n = len(C)
    p = len(A) - len(C)
    if not (len(A) == len(B) == len(C)+p == len(D)+p):
        if verbose:
            print(f'Base sequences should have length n+p, n+p, n, n, found {len(A)}, {len(B)}, {len(C)}, {len(D)}')
        return False

    for seq in base_sequences:
        for el in seq:
            if abs(el) != 1:
                if verbose:
                    print(f'Base sequences should only contain -1, +1, found {el}')
                return False

    for j in range(1, n+p):
        autocorr = _nonperiodic_autocorrelation(A, j) + _nonperiodic_autocorrelation(B, j) + _nonperiodic_autocorrelation(C, j) + _nonperiodic_autocorrelation(D, j)
        if autocorr != 0:
            if verbose:
                print(f"Nonperiodic autocorrelation should always be zero, found {autocorr} for parameter {j}")
            return False

    return True


def turyn_type_sequences_smallcases(n, existence=False):
    r"""
    Construction of Turyn type sequences for small values of `n`.

    The data is taken from [KTR2005]_ for `n= 36`, and from [BDKR2013]_ for `n\le 32`.

    INPUT:

    - ``n`` -- integer; the length of the Turyn type sequences

    - ``existence`` -- boolean (default: ``False``); if ``True``, only return
      whether the Turyn type sequences are available for the given length

    EXAMPLES:

    By default, this method returns the four Turyn type sequences ::

        sage: from sage.combinat.t_sequences import turyn_type_sequences_smallcases
        sage: turyn_type_sequences_smallcases(4)
        [[1, 1, 1, 1], [1, 1, -1, 1], [1, 1, -1, -1], [1, -1, 1]]

    If we pass the ``existence`` flag, the method will return a boolean ::

        sage: turyn_type_sequences_smallcases(4, existence=True)
        True

    TESTS::

        sage: turyn_type_sequences_smallcases(17)
        Traceback (most recent call last):
        ...
        ValueError: Turyn type sequences of length 17 are not implemented yet.
        sage: turyn_type_sequences_smallcases(17, existence=True)
        False

    ALGORITHM:

    The Turyn type sequences are stored in hexadecimal format.
    Given `n` hexadecimal digits `h_1, h_2,...,h_n`, it is possible to get the
    Turyn type sequences by converting each `h_i` (`1 \le i \le n-1`) into a
    four digits binary number. Then, the `j`-th binary digit is `0` if the `i`-th
    number in the `j`-th sequence is `1`, and it is `1` if the number in the
    sequence is -1.

    For the `n`-th digit, it should be converted to a 3 digits binary number, and
    then the same mapping as before can be used (see also [BDKR2013]_).
    """
    def convertLists(hexstring):
        seqs = [Sequence([]), Sequence([]), Sequence([]), Sequence([])]
        for c in hexstring[:-1]:
            binary = bin(int(c, 16))[2:].zfill(4)
            for i in range(4):
                if binary[i] == '0':
                    seqs[i].append(1)
                else:
                    seqs[i].append(-1)
        last = bin(int(hexstring[-1], 16))[2:].zfill(3)
        for i in range(3):
            if last[i] == '0':
                seqs[i].append(1)
            else:
                seqs[i].append(-1)
        return seqs

    db = {
        2: '01',
        4: '0161',
        6: '006d61',
        8: '06e5c4d1',
        10: '0001f4a961',
        12: '0004f90bc961',
        14: '00036ac71c7651',
        16: '0000778e52de5561',
        18: '00006758b30d1e9a51',
        20: '000038e2739c7a0b6951',
        22: '00000f702c71a9ad565961',
        24: '00000b7c2cb2bc4b6cd9a961',
        26: '000000ff0f846f1ca5a5aa9551',
        28: '0000067cde3e50639ab46135aa51',
        30: '000000f70b106f9d427a25e9a96951',
        32: '00000138f64f1c1e77844f26d95a5961',
        36: '060989975b685d8fc80750b21c0212eceb26',
    }

    if existence:
        return n in db

    if n not in db:
        raise ValueError(f"Turyn type sequences of length {n} are not implemented yet.")

    return convertLists(db[n])


def base_sequences_smallcases(n, p, existence=False, check=True):
    r"""Construct base sequences of length `n+p, n+p, n, n` from available data.

    The function uses the construction :func:`base_sequences_construction`,
    together with Turyn type sequences from :func:`turyn_type_sequences_smallcases`
    to construct base sequences with `p = n-1`.

    Furthermore, this function uses also Turyn sequences (i.e. base sequences
    with `p=1`) from :func:`turyn_sequences_smallcases`.

    INPUT:

    - ``n`` -- integer; the length of the last two base sequences

    - ``p`` -- integer; `n+p` will be the length of the first two base
      sequences

    - ``existence`` -- boolean (default: ``False``); if ``True``, the function
      will only check whether the base sequences can be constructed

    - ``check`` -- boolean (default: ``True``); check that the resulting
      sequences are base sequences before returning them

    OUTPUT:

    If ``existence`` is ``False``, the function returns a list containing the
    four base sequences, or raises an error if the base sequences cannot be
    constructed. If ``existence`` is ``True``, the function returns a boolean,
    which is ``True`` if the base sequences can be constructed and ``False``
    otherwise.

    EXAMPLES::

        sage: from sage.combinat.t_sequences import base_sequences_smallcases
        sage: base_sequences_smallcases(8, 7)
        [[1, -1, -1, 1, 1, 1, 1, -1, 1, 1, 1, -1, 1, 1, -1],
        [1, -1, -1, 1, 1, 1, 1, -1, -1, -1, -1, 1, -1, -1, 1],
        [1, 1, -1, 1, -1, 1, -1, 1],
        [1, -1, -1, -1, -1, -1, -1, 1]]

    If ``existence`` is ``True``, the function returns a boolean ::

        sage: base_sequences_smallcases(8, 7, existence=True)
        True
        sage: base_sequences_smallcases(7, 5, existence=True)
        False

    TESTS::

        sage: base_sequences_smallcases(7, 5)
        Traceback (most recent call last):
        ...
        ValueError: Base sequences of order 12, 12, 7, 7 not yet implemented.
        sage: seqs = base_sequences_smallcases(16, 15)
        sage: len(seqs[0]) == len(seqs[1]) == 16+15
        True
        sage: len(seqs[2]) == len(seqs[3]) == 16
        True
    """

    if existence:
        return p == n-1 and turyn_type_sequences_smallcases(n, existence=True)

    if p == n-1 and turyn_type_sequences_smallcases(n, existence=True):
        if existence:
            return True
        turyn_type_seqs = turyn_type_sequences_smallcases(n)
        return base_sequences_construction(turyn_type_seqs, check=check)
    if p == 1 and turyn_sequences_smallcases(n + p, existence=True):
        if existence:
            return True
        return turyn_sequences_smallcases(n + p)

    raise ValueError(f'Base sequences of order {n+p}, {n+p}, {n}, {n} not yet implemented.')



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.rings.polynomial.pbori
"""
An ANF to CNF Converter using a Dense/Sparse Strategy

This converter is based on two converters. The first one, by Martin Albrecht, was based on [CB2007]_,
this is the basis of the "dense" part of the converter. It was later improved by Mate Soos. The
second one, by Michael Brickenstein, uses a reduced truth table based approach and forms the
"sparse" part of the converter.

AUTHORS:

- Martin Albrecht - (2008-09) initial version of 'anf2cnf.py'
- Michael Brickenstein - (2009) 'cnf.py' for PolyBoRi
- Mate Soos - (2010) improved version of 'anf2cnf.py'
- Martin Albrecht - (2012) unified and added to Sage

Classes and Methods
-------------------
"""

# ###########################################################################
#  Copyright (C) 2008-2009 Martin Albrecht <martinralbrecht@googlemail.com>
#  Copyright (C) 2009 Michael Brickenstein <brickenstein@mfo.de>
#  Copyright (C) 2010 Mate Soos
#  Copyright (C) 2012 Martin Albrecht <martinralbrecht@googlemail.com>
#  Distributed under the terms of the GNU General Public License (GPL)
#  The full text of the GPL is available at:
#                  https://www.gnu.org/licenses/
# ###########################################################################

from random import Random

from sage.rings.integer_ring import ZZ
from sage.functions.other import ceil
from sage.misc.cachefunc import cached_method, cached_function
from sage.misc.lazy_import import lazy_import
from sage.combinat.permutation import Permutations
from sage.sat.converters import ANF2CNFConverter

lazy_import('sage.rings.polynomial.pbori.pbori', 'if_then_else', as_='ite')


class CNFEncoder(ANF2CNFConverter):
    """
    ANF to CNF Converter using a Dense/Sparse Strategy. This converter distinguishes two classes of
    polynomials.

    1. Sparse polynomials are those with at most ``max_vars_sparse`` variables. Those are converted
    using reduced truth-tables based on PolyBoRi's internal representation.

    2. Polynomials with more variables are converted by introducing new variables for monomials and
    by converting these linearised polynomials.

    Linearised polynomials are converted either by splitting XOR chains -- into chunks of length
    ``cutting_number`` -- or by constructing XOR clauses if the underlying solver supports it. This
    behaviour is disabled by passing ``use_xor_clauses=False``.

    .. automethod:: __init__
    .. automethod:: __call__
    """
    def __init__(self, solver, ring, max_vars_sparse=6, use_xor_clauses=None, cutting_number=6, random_seed=16):
        """
        Construct ANF to CNF converter over ``ring`` passing clauses to ``solver``.

        INPUT:

        - ``solver`` -- a SAT-solver instance

        - ``ring`` -- a :class:`sage.rings.polynomial.pbori.BooleanPolynomialRing`

        - ``max_vars_sparse`` -- maximum number of variables for direct conversion

        - ``use_xor_clauses`` -- use XOR clauses; if ``None`` use if
          ``solver`` supports it. (default: ``None``)

        - ``cutting_number`` -- maximum length of XOR chains after
          splitting if XOR clauses are not supported (default: 6)

        - ``random_seed`` -- the direct conversion method uses
          randomness, this sets the seed (default: 16)

        EXAMPLES:

        We compare the sparse and the dense strategies, sparse first::

            sage: B.<a,b,c> = BooleanPolynomialRing()
            sage: from sage.sat.converters.polybori import CNFEncoder
            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: fn = tmp_filename()
            sage: solver = DIMACS(filename=fn)
            sage: e = CNFEncoder(solver, B)
            sage: e.clauses_sparse(a*b + a + 1)
            sage: _ = solver.write()
            sage: print(open(fn).read())
            p cnf 3 2
            -2 0
            1 0
            sage: e.phi
            [None, a, b, c]

        Now, we convert using the dense strategy::

            sage: B.<a,b,c> = BooleanPolynomialRing()
            sage: from sage.sat.converters.polybori import CNFEncoder
            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: fn = tmp_filename()
            sage: solver = DIMACS(filename=fn)
            sage: e = CNFEncoder(solver, B)
            sage: e.clauses_dense(a*b + a + 1)
            sage: _ = solver.write()
            sage: print(open(fn).read())
            p cnf 4 5
            1 -4 0
            2 -4 0
            4 -1 -2 0
            -4 -1 0
            4 1 0
            sage: e.phi
            [None, a, b, c, a*b]

        .. NOTE::

            This constructor generates SAT variables for each Boolean polynomial variable.
        """
        self.random_generator = Random(random_seed)
        self.one_set = ring.one().set()
        self.empty_set = ring.zero().set()

        self.solver = solver
        self.max_vars_sparse = max_vars_sparse
        self.cutting_number = cutting_number

        if use_xor_clauses is None:
            use_xor_clauses = hasattr(solver, "add_xor_clause")
        self.use_xor_clauses = use_xor_clauses

        self.ring = ring

        # If you change this, make sure we are calling m.index()
        # below, as this relies on phi being sorted like this.
        self._phi = [None]
        for x in sorted([x.lm() for x in self.ring.gens()], key=lambda x: x.index()):
            self.var(x)

    def var(self, m=None, decision=None):
        """
        Return a *new* variable.

        This is a thin wrapper around the SAT-solvers function where
        we keep track of which SAT variable corresponds to which
        monomial.

        INPUT:

        - ``m`` -- something the new variables maps to, usually a monomial
        - ``decision`` -- is this variable a decision variable?

        EXAMPLES::

            sage: from sage.sat.converters.polybori import CNFEncoder
            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: B.<a,b,c> = BooleanPolynomialRing()
            sage: ce = CNFEncoder(DIMACS(), B)
            sage: ce.var()
            4
        """
        self._phi.append(m)
        return self.solver.var(decision=decision)

    @property
    def phi(self):
        """
        Map SAT variables to polynomial variables.

        EXAMPLES::

            sage: from sage.sat.converters.polybori import CNFEncoder
            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: B.<a,b,c> = BooleanPolynomialRing()
            sage: ce = CNFEncoder(DIMACS(), B)
            sage: ce.var()
            4
            sage: ce.phi
            [None, a, b, c, None]
        """
        return list(self._phi)

    ##################################################
    # Encoding based on polynomial roots
    ##################################################

    def zero_blocks(self, f):
        """
        Divide the zero set of ``f`` into blocks.

        EXAMPLES::

            sage: B.<a,b,c> = BooleanPolynomialRing()
            sage: from sage.sat.converters.polybori import CNFEncoder
            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: e = CNFEncoder(DIMACS(), B)
            sage: sorted(sorted(d.items()) for d in e.zero_blocks(a*b*c))
            [[(c, 0)], [(b, 0)], [(a, 0)]]

        .. NOTE::

            This function is randomised.
        """
        variables = f.vars_as_monomial()

        space = variables.divisors()
        variables = list(variables.variables())
        zeros = f.zeros_in(space)
        rest = zeros
        res = list()

        def choose(s):
            indices = []
            assert not s.empty()
            nav = s.navigation()
            while not nav.constant():
                e = nav.else_branch()
                t = nav.then_branch()
                if e.constant() and not e.terminal_one():
                    indices.append(nav.value())
                    nav = t
                else:
                    if self.random_generator.randint(0, 1):
                        indices.append(nav.value())
                        nav = t

                    else:
                        nav = e
            assert nav.terminal_one()
            res = self.one_set
            for i in reversed(indices):
                res = ite(i, res, self.empty_set)
            return next(iter(res))

        while not rest.empty():
            l = choose(rest)
            l_variables = set(l.variables())
            block_dict = {v: (1 if v in l_variables else 0) for v in variables}
            l = l.set()
            self.random_generator.shuffle(variables)
            for v in variables:
                candidate = l.change(v.index())
                if candidate.diff(zeros).empty():
                    l = l.union(candidate)
                    del block_dict[v]
            rest = rest.diff(l)
            res.append(block_dict)
        return res

    def clauses_sparse(self, f):
        """
        Convert ``f`` using the sparse strategy.

        INPUT:

        - ``f`` -- a :class:`sage.rings.polynomial.pbori.BooleanPolynomial`

        EXAMPLES::

            sage: B.<a,b,c> = BooleanPolynomialRing()
            sage: from sage.sat.converters.polybori import CNFEncoder
            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: fn = tmp_filename()
            sage: solver = DIMACS(filename=fn)
            sage: e = CNFEncoder(solver, B)
            sage: e.clauses_sparse(a*b + a + 1)
            sage: _ = solver.write()
            sage: print(open(fn).read())
            p cnf 3 2
            -2 0
            1 0
            sage: e.phi
            [None, a, b, c]
        """
        # we form an expression for a var configuration *not* lying in
        # the block it is evaluated to 0 by f, iff it is not lying in
        # any zero block of f+1

        blocks = self.zero_blocks(f + 1)
        C = [{variable: 1 - value for variable, value in b.items()}
             for b in blocks]

        def to_dimacs_index(v):
            return v.index() + 1

        def clause(c):
            return [to_dimacs_index(variable)
                    if value == 1 else -to_dimacs_index(variable)
                    for variable, value in c.items()]

        data = (clause(c) for c in C)
        for d in sorted(data):
            self.solver.add_clause(d)

    ###################################################
    # Indirect conversion, may add new variables
    ###################################################

    def clauses_dense(self, f):
        """
        Convert ``f`` using the dense strategy.

        INPUT:

        - ``f`` -- a :class:`sage.rings.polynomial.pbori.BooleanPolynomial`

        EXAMPLES::

            sage: B.<a,b,c> = BooleanPolynomialRing()
            sage: from sage.sat.converters.polybori import CNFEncoder
            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: fn = tmp_filename()
            sage: solver = DIMACS(filename=fn)
            sage: e = CNFEncoder(solver, B)
            sage: e.clauses_dense(a*b + a + 1)
            sage: _ = solver.write()
            sage: print(open(fn).read())
            p cnf 4 5
            1 -4 0
            2 -4 0
            4 -1 -2 0
            -4 -1 0
            4 1 0
            sage: e.phi
            [None, a, b, c, a*b]
        """
        equal_zero = not bool(f.constant_coefficient())

        f = (f - f.constant_coefficient())
        f = [self.monomial(m) for m in f]

        if self.use_xor_clauses:
            self.solver.add_xor_clause(f, rhs=not equal_zero)
        elif len(f) > self.cutting_number:
            for fpart, this_equal_zero in self.split_xor(f, equal_zero):
                ll = len(fpart)
                for p in self.permutations(ll, this_equal_zero):
                    self.solver.add_clause([p[i] * fpart[i] for i in range(ll)])
        else:
            ll = len(f)
            for p in self.permutations(ll, equal_zero):
                self.solver.add_clause([p[i] * f[i] for i in range(ll)])

    @cached_method
    def monomial(self, m):
        """
        Return SAT variable for ``m``.

        INPUT:

        - ``m`` -- a monomial

        OUTPUT: an index for a SAT variable corresponding to ``m``

        EXAMPLES::

            sage: B.<a,b,c> = BooleanPolynomialRing()
            sage: from sage.sat.converters.polybori import CNFEncoder
            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: fn = tmp_filename()
            sage: solver = DIMACS(filename=fn)
            sage: e = CNFEncoder(solver, B)
            sage: e.clauses_dense(a*b + a + 1)
            sage: e.phi
            [None, a, b, c, a*b]

        If monomial is called on a new monomial, a new variable is created::

            sage: e.monomial(a*b*c)
            5
            sage: e.phi
            [None, a, b, c, a*b, a*b*c]

        If monomial is called on a monomial that was queried before,
        the index of the old variable is returned and no new variable
        is created::

            sage: e.monomial(a*b)
            4
            sage: e.phi
            [None, a, b, c, a*b, a*b*c]

        .. NOTE::

            For correctness, this function is cached.
        """
        if m.deg() == 1:
            return m.index() + 1

        # we need to encode the relationship between the monomial
        # and its variables
        variables = [self.monomial(v) for v in m.variables()]
        monomial = self.var(m)

        # (a | -w) & (b | -w) & (w | -a | -b) <=> w == a*b
        for v in variables:
            self.solver.add_clause((v, -monomial))
        self.solver.add_clause(tuple([monomial] + [-v for v in variables]))

        return monomial

    @cached_function
    def permutations(length, equal_zero):
        """
        Return permutations of length ``length`` which are equal to
        zero if ``equal_zero`` and equal to one otherwise.

        A variable is false if the integer in its position is smaller
        than zero and true otherwise.

        INPUT:

        - ``length`` -- the number of variables
        - ``equal_zero`` -- should the sum be equal to zero?

        EXAMPLES::

            sage: from sage.sat.converters.polybori import CNFEncoder
            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: B.<a,b,c> = BooleanPolynomialRing()
            sage: ce = CNFEncoder(DIMACS(), B)
            sage: ce.permutations(3, True)
            [[-1, -1, -1], [1, 1, -1], [1, -1, 1], [-1, 1, 1]]

            sage: ce.permutations(3, False)
            [[1, -1, -1], [-1, 1, -1], [-1, -1, 1], [1, 1, 1]]
        """
        E = []
        for num_negated in range(length + 1):
            if ((num_negated % 2) ^ ((length + 1) % 2)) == equal_zero:
                continue
            start = [1 for _ in range(num_negated)]
            start.extend(-1 for _ in range(length - num_negated))
            E.extend(Permutations(start))
        return E

    def split_xor(self, monomial_list, equal_zero):
        """
        Split XOR chains into subchains.

        INPUT:

        - ``monomial_list`` -- list of monomials
        - ``equal_zero`` -- is the constant coefficient zero?

        EXAMPLES::

            sage: from sage.sat.converters.polybori import CNFEncoder
            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: B.<a,b,c,d,e,f> = BooleanPolynomialRing()
            sage: ce = CNFEncoder(DIMACS(), B, cutting_number=3)
            sage: ce.split_xor([1,2,3,4,5,6], False)
            [[[1, 7], False], [[7, 2, 8], True], [[8, 3, 9], True], [[9, 4, 10], True], [[10, 5, 11], True], [[11, 6], True]]

            sage: ce = CNFEncoder(DIMACS(), B, cutting_number=4)
            sage: ce.split_xor([1,2,3,4,5,6], False)
            [[[1, 2, 7], False], [[7, 3, 4, 8], True], [[8, 5, 6], True]]

            sage: ce = CNFEncoder(DIMACS(), B, cutting_number=5)
            sage: ce.split_xor([1,2,3,4,5,6], False)
            [[[1, 2, 3, 7], False], [[7, 4, 5, 6], True]]
        """
        c = self.cutting_number

        nm = len(monomial_list)
        step = ceil((c-2)/ZZ(nm) * nm)
        M = []

        new_variables = []
        for j in range(0, nm, step):
            m = new_variables + monomial_list[j:j+step]
            if (j + step) < nm:
                new_variables = [self.var(None)]
                m += new_variables
            M.append([m, equal_zero])
            equal_zero = True
        return M

    ###################################################
    # Highlevel Functions
    ###################################################

    def clauses(self, f):
        """
        Convert ``f`` using the sparse strategy if ``f.nvariables()`` is
        at most ``max_vars_sparse`` and the dense strategy otherwise.

        INPUT:

        - ``f`` -- a :class:`sage.rings.polynomial.pbori.BooleanPolynomial`

        EXAMPLES::

            sage: B.<a,b,c> = BooleanPolynomialRing()
            sage: from sage.sat.converters.polybori import CNFEncoder
            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: fn = tmp_filename()
            sage: solver = DIMACS(filename=fn)
            sage: e = CNFEncoder(solver, B, max_vars_sparse=2)
            sage: e.clauses(a*b + a + 1)
            sage: _ = solver.write()
            sage: print(open(fn).read())
            p cnf 3 2
            -2 0
            1 0
            sage: e.phi
            [None, a, b, c]

            sage: B.<a,b,c> = BooleanPolynomialRing()
            sage: from sage.sat.converters.polybori import CNFEncoder
            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: fn = tmp_filename()
            sage: solver = DIMACS(filename=fn)
            sage: e = CNFEncoder(solver, B, max_vars_sparse=2)
            sage: e.clauses(a*b + a + c)
            sage: _ = solver.write()
            sage: print(open(fn).read())
            p cnf 4 7
            1 -4 0
            2 -4 0
            4 -1 -2 0
            -4 -1 -3 0
            4 1 -3 0
            4 -1 3 0
            -4 1 3 0

            sage: e.phi
            [None, a, b, c, a*b]
        """
        if f.nvariables() <= self.max_vars_sparse:
            self.clauses_sparse(f)
        else:
            self.clauses_dense(f)

    def __call__(self, F):
        """
        Encode the boolean polynomials in ``F`` .

        INPUT:

        - ``F`` -- an iterable of :class:`sage.rings.polynomial.pbori.BooleanPolynomial`

        OUTPUT: an inverse map int -> variable

        EXAMPLES::

            sage: B.<a,b,c> = BooleanPolynomialRing()
            sage: from sage.sat.converters.polybori import CNFEncoder
            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: fn = tmp_filename()
            sage: solver = DIMACS(filename=fn)
            sage: e = CNFEncoder(solver, B, max_vars_sparse=2)
            sage: e([a*b + a + 1, a*b+ a + c])
            [None, a, b, c, a*b]
            sage: _ = solver.write()
            sage: print(open(fn).read())
            p cnf 4 9
            -2 0
            1 0
            1 -4 0
            2 -4 0
            4 -1 -2 0
            -4 -1 -3 0
            4 1 -3 0
            4 -1 3 0
            -4 1 3 0

            sage: e.phi
            [None, a, b, c, a*b]
        """
        for f in F:
            self.clauses(f)
        return self.phi

    ####################################################
    # Highlevel Functions
    ###################################################

    def to_polynomial(self, c):
        """
        Convert clause to :class:`sage.rings.polynomial.pbori.BooleanPolynomial`.

        INPUT:

        - ``c`` -- a clause

        EXAMPLES::

            sage: B.<a,b,c> = BooleanPolynomialRing()
            sage: from sage.sat.converters.polybori import CNFEncoder
            sage: from sage.sat.solvers.dimacs import DIMACS
            sage: fn = tmp_filename()
            sage: solver = DIMACS(filename=fn)
            sage: e = CNFEncoder(solver, B, max_vars_sparse=2)
            sage: _ = e([a*b + a + 1, a*b+ a + c])
            sage: e.to_polynomial( (1,-2,3) )
            a*b*c + a*b + b*c + b
        """
        phi = self.phi
        product = self.ring.one()
        for v in c:
            if phi[abs(v)] is None:
                raise ValueError("clause contains an XOR glueing variable")
            product *= phi[abs(v)] + int(v > 0)
        return product



# sage_setup: distribution = sagemath-combinat
r"""
Species structures

We will illustrate the use of the structure classes using the
"balls and bars" model for integer compositions. An integer
composition of 6 such as [2, 1, 3] can be represented in this model
as 'oooooo' where the 6 o's correspond to the balls and the 2 's
correspond to the bars. If BB is our species for this model, the it
satisfies the following recursive definition:

BB = o + o\*BB + o\*|\*BB

Here we define this species using the default structures::

    sage: ball = species.SingletonSpecies()
    sage: bar = species.EmptySetSpecies()
    sage: BB = CombinatorialSpecies()
    sage: BB.define(ball + ball*BB + ball*bar*BB)
    sage: o = var('o')                                                                  # needs sage.symbolic
    sage: BB.isotypes([o]*3).list()                                                     # needs sage.symbolic
    [o*(o*o), o*((o*{})*o), (o*{})*(o*o), (o*{})*((o*{})*o)]

If we ignore the parentheses, we can read off that the integer
compositions are [3], [2, 1], [1, 2], and [1, 1, 1].
"""
#*****************************************************************************
#       Copyright (C) 2008 Mike Hansen <mhansen@gmail.com>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************
from sage.categories.enumerated_sets import EnumeratedSets
from sage.combinat.combinat import CombinatorialObject
from sage.rings.integer import Integer
from sage.structure.parent import Parent
from copy import copy


class GenericSpeciesStructure(CombinatorialObject):
    def __init__(self, parent, labels, list):
        """
        This is a base class from which the classes for the structures inherit.

        EXAMPLES::

            sage: from sage.combinat.species.structure import GenericSpeciesStructure
            sage: a = GenericSpeciesStructure(None, [2,3,4], [1,2,3])
            sage: a
            [2, 3, 4]
            sage: a.parent() is None
            True
            sage: a == loads(dumps(a))
            True
        """
        self._parent = parent
        self._labels = labels
        CombinatorialObject.__init__(self, list)

    def parent(self):
        """
        Return the species that this structure is associated with.

        EXAMPLES::

            sage: L = species.LinearOrderSpecies()
            sage: a,b = L.structures([1,2])
            sage: a.parent()
            Linear order species
        """
        try:
            return self._parent
        except AttributeError:
            raise NotImplementedError

    def __repr__(self):
        """
        EXAMPLES::

            sage: from sage.combinat.species.structure import GenericSpeciesStructure
            sage: a = GenericSpeciesStructure(None, [2,3,4], [1,2,3])
            sage: a
            [2, 3, 4]
        """
        return repr([self._relabel(i) for i in self._list])

    def __eq__(self, other):
        """
        Check whether ``self`` is equal to ``other``.

        EXAMPLES::

            sage: T = species.BinaryTreeSpecies()
            sage: t = T.structures([1,2,3])[0]; t
            1*(2*3)
            sage: t[0], t[1][0]
            (1, 2)
            sage: t[0] == t[1][0]
            False
        """
        if not isinstance(other, GenericSpeciesStructure):
            return False
        return self._list == other._list and self.labels() == other.labels()

    def __ne__(self, other):
        """
        Check whether ``self`` is not equal to ``other``.

        EXAMPLES::

            sage: T = species.BinaryTreeSpecies()
            sage: t = T.structures([1,2,3])[0]; t
            1*(2*3)
            sage: t[0], t[1][0]
            (1, 2)
            sage: t[0] != t[1][0]
            True
        """
        return not (self == other)

    def labels(self):
        """
        Return the labels used for this structure.

        .. NOTE::

            This includes labels which may not "appear" in this
            particular structure.

        EXAMPLES::

            sage: P = species.SubsetSpecies()
            sage: s = P.structures(["a", "b", "c"]).random_element()
            sage: s.labels()
            ['a', 'b', 'c']
        """
        return copy(self._labels)

    def change_labels(self, labels):
        """
        Return a relabelled structure.

        INPUT:

        - ``labels`` -- list of labels

        OUTPUT:

        A structure with the `i`-th label of ``self`` replaced with the `i`-th
        label of the list.

        EXAMPLES::

            sage: P = species.SubsetSpecies()
            sage: S = P.structures(["a", "b", "c"])
            sage: [s.change_labels([1,2,3]) for s in S]
            [{}, {1}, {2}, {3}, {1, 2}, {1, 3}, {2, 3}, {1, 2, 3}]
        """
        c = copy(self)
        c._labels = labels
        return c

    def _relabel(self, i):
        """
        EXAMPLES::

            sage: from sage.combinat.species.structure import GenericSpeciesStructure
            sage: a = GenericSpeciesStructure(None, [2,3,4], [1,2,3])
            sage: a._relabel(1)
            2
            sage: a._relabel([1,2,3])
            [1, 2, 3]
        """
        if isinstance(i, (int, Integer)):
            return self._labels[i-1]
        else:
            return i

    def is_isomorphic(self, x):
        """
        EXAMPLES::

            sage: S = species.SetSpecies()
            sage: a = S.structures([1,2,3]).random_element(); a
            {1, 2, 3}
            sage: b = S.structures(['a','b','c']).random_element(); b
            {'a', 'b', 'c'}
            sage: a.is_isomorphic(b)
            True
        """
        if self.__class__ != x.__class__:
            return False
        if self.parent() != x.parent():
            return False

        #We don't care about the labels for isomorphism testing
        if self.canonical_label()._list == x.canonical_label()._list:
            return True
        else:
            return False


#For backward compatibility.  This should be removed in the near
#future since I doubt that there is any code that depends directly on
#SpeciesStructure.
SpeciesStructure = GenericSpeciesStructure


class SpeciesStructureWrapper(GenericSpeciesStructure):
    def __init__(self, parent, s, **options):
        """
        This is a class for the structures of species such as the sum
        species that do not provide "additional" structure.  For example,
        if you have the sum `C` of species `A` and `B`,
        then a structure of `C` will either be either something from `A` or `B`.
        Instead of just returning one of these directly, a "wrapper" is
        put around them so that they have their parent is `C` rather than `A` or
        `B`::

            sage: X = species.SingletonSpecies()
            sage: X2 = X+X
            sage: s = X2.structures([1]).random_element(); s
            1
            sage: s.parent()
            Sum of (Singleton species) and (Singleton species)
            sage: from sage.combinat.species.structure import SpeciesStructureWrapper
            sage: issubclass(type(s), SpeciesStructureWrapper)
            True

        EXAMPLES::

            sage: E = species.SetSpecies(); B = E+E
            sage: s = B.structures([1,2,3]).random_element()
            sage: s.parent()
            Sum of (Set species) and (Set species)
            sage: s == loads(dumps(s))
            True
        """
        self._parent = parent
        self._s = s
        self._options = options
        GenericSpeciesStructure.__init__(self, parent, s._labels, s._list)

    def __getattr__(self, attr):
        """
        EXAMPLES::

            sage: E = species.SetSpecies(); B = E+E
            sage: s = B.structures([1,2,3]).random_element()
            sage: s
            {1, 2, 3}
        """
        if attr == "_s":
            return None

        return getattr(self._s, attr)

    def __repr__(self):
        """
        Return the repr of the object which this one wraps.

        EXAMPLES::

            sage: E = species.SetSpecies()
            sage: s = (E+E).structures([1,2,3]).random_element(); s
            {1, 2, 3}
        """
        return repr(self._s)

    def transport(self, perm):
        """
        EXAMPLES::

            sage: P = species.PartitionSpecies()
            sage: s = (P+P).structures([1,2,3])[1]; s                                   # needs sage.libs.flint
            {{1, 3}, {2}}
            sage: s.transport(PermutationGroupElement((2,3)))                           # needs sage.groups sage.libs.flint
            {{1, 2}, {3}}
        """
        return self.__class__(self._parent, self._s.transport(perm), **self._options)

    def canonical_label(self):
        """
        EXAMPLES::

            sage: P = species.PartitionSpecies()
            sage: s = (P+P).structures([1,2,3])[1]; s                                   # needs sage.libs.flint
            {{1, 3}, {2}}
            sage: s.canonical_label()                                                   # needs sage.libs.flint
            {{1, 2}, {3}}
        """
        return self.__class__(self._parent, self._s.canonical_label(), **self._options)

    def change_labels(self, labels):
        """
        Return a relabelled structure.

        INPUT:

        - ``labels`` -- list of labels

        OUTPUT:

        A structure with the `i`-th label of ``self`` replaced with the `i`-th
        label of the list.

        EXAMPLES::

            sage: X = species.SingletonSpecies()
            sage: X2 = X+X
            sage: s = X2.structures([1]).random_element(); s
            1
            sage: s.change_labels(['a'])
            'a'
        """
        c = GenericSpeciesStructure.change_labels(self, labels)
        c._s = c._s.change_labels(labels)
        return c


##############################################################


class SpeciesWrapper(Parent):
    def __init__(self, species, labels, iterator, generating_series, name, structure_class):
        """
        This is a abstract base class for the set of structures of a
        species as well as the set of isotypes of the species.

        .. NOTE::

            One typically does not use :class:`SpeciesWrapper`
            directly, but instead instantiates one of its subclasses:
            :class:`StructuresWrapper` or :class:`IsotypesWrapper`.

        EXAMPLES::

            sage: from sage.combinat.species.structure import SpeciesWrapper
            sage: F = species.SetSpecies()
            sage: S = SpeciesWrapper(F, [1,2,3], "_structures", "generating_series", 'Structures', None)
            sage: S
            Structures for Set species with labels [1, 2, 3]
            sage: S.list()
            [{1, 2, 3}]
            sage: S.cardinality()
            1
        """
        Parent.__init__(self, category=EnumeratedSets().Finite())
        self._species = species
        self._labels = labels
        self._iterator = iterator
        self._generating_series = generating_series
        self._name = "%s for %s with labels %s" % (name, species, labels)
        self._structure_class = structure_class if structure_class is not None else species._default_structure_class

    def __eq__(self, other) -> bool:
        r"""
        EXAMPLES::

            sage: from sage.combinat.species.structure import SpeciesWrapper
            sage: F = species.SetSpecies()
            sage: S = SpeciesWrapper(F, [1,2,3], "_structures", "generating_series", 'Structures', None)
            sage: S == SpeciesWrapper(F, [1,2,3], "_structures", "generating_series", 'Structures', None)
            True
        """
        return ((self._species, self._labels,
                 self._iterator, self._generating_series,
                 self._name, self._structure_class) == (other._species, other._labels,
                                                        other._iterator, other._generating_series,
                                                        other._name, other._structure_class))

    def __ne__(self, other) -> bool:
        r"""
        EXAMPLES::

            sage: from sage.combinat.species.structure import SpeciesWrapper
            sage: F = species.SetSpecies()
            sage: S = SpeciesWrapper(F, [1,2,3], "_structures", "generating_series", 'Structures', None)
            sage: S != SpeciesWrapper(F, [1,2,3], "_structures", "generating_series", 'Structures', None)
            False
        """
        return not (self == other)

    def _repr_(self) -> str:
        """
        EXAMPLES::

            sage: from sage.combinat.species.structure import SpeciesWrapper
            sage: F = species.SetSpecies()
            sage: S = SpeciesWrapper(F, [1,2,3], "_structures", "generating_series", 'Structures', None)
            sage: repr(S)   # indirect doctest
            'Structures for Set species with labels [1, 2, 3]'
        """
        return self._name

    def labels(self):
        """
        Return the labels used on these structures.  If `X` is the
        species, then :meth:`labels` returns the preimage of these
        structures under the functor `X`.

        EXAMPLES::

            sage: F = species.SetSpecies()
            sage: F.structures([1,2,3]).labels()
            [1, 2, 3]
        """
        return copy(self._labels)

    def __iter__(self):
        """
        EXAMPLES::

            sage: F = species.SetSpecies()
            sage: F.structures([1,2,3]).list()
            [{1, 2, 3}]
        """
        #If the min and max are set, then we want to make sure
        #that the iterator respects those bounds.
        if (self._species._min is not None and
            len(self._labels) < self._species._min):
            return iter([])

        if (self._species._max is not None and
            len(self._labels) >= self._species._max):
            return iter([])

        #We check to see if the
        try:
            if self.cardinality() == 0:
                return iter([])
        except TypeError:
            raise NotImplementedError

        return getattr(self._species, self._iterator)(self._structure_class, self._labels)

    def cardinality(self):
        """
        Return the number of structures in this set.

        EXAMPLES::

            sage: F = species.SetSpecies()
            sage: F.structures([1,2,3]).cardinality()
            1
        """
        return getattr(self._species, self._generating_series)().count(len(self._labels))


class StructuresWrapper(SpeciesWrapper):
    def __init__(self, species, labels, structure_class):
        """
        A base class for the set of structures of a species with given
        set of labels.  An object of this type is returned when you
        call the :meth:`structures` method of a species.

        EXAMPLES::

            sage: F = species.SetSpecies()
            sage: S = F.structures([1,2,3])
            sage: S == loads(dumps(S))
            True
        """
        SpeciesWrapper.__init__(self, species, labels,
                                "_structures",
                                "generating_series",
                                "Structures",
                                structure_class)


class IsotypesWrapper(SpeciesWrapper):
    def __init__(self, species, labels, structure_class):
        """
        A base class for the set of isotypes of a species with given
        set of labels.  An object of this type is returned when you
        call the :meth:`isotypes` method of a species.

        EXAMPLES::

            sage: F = species.SetSpecies()
            sage: S = F.isotypes([1,2,3])
            sage: S == loads(dumps(S))
            True
        """
        SpeciesWrapper.__init__(self, species, labels,
                                "_isotypes",
                                "isotype_generating_series",
                                "Isomorphism types",
                                structure_class)


class SimpleStructuresWrapper(SpeciesWrapper):
    def __init__(self, species, labels, structure_class):
        """
        .. warning::

            This is deprecated and currently not used for anything.

        EXAMPLES::

            sage: F = species.SetSpecies()
            sage: S = F.structures([1,2,3])
            sage: S == loads(dumps(S))
            True
        """
        SpeciesWrapper.__init__(self, species, labels,
                                "_simple_structures_selector",
                                "generating_series",
                                "Simple structures",
                                structure_class)


class SimpleIsotypesWrapper(SpeciesWrapper):
    def __init__(self, species, labels, structure_class):
        """
        .. warning::

            This is deprecated and currently not used for anything.

        EXAMPLES::

            sage: F = species.SetSpecies()
            sage: S = F.structures([1,2,3])
            sage: S == loads(dumps(S))
            True
        """
        SpeciesWrapper.__init__(self, species, labels,
                                "_simple_isotypes_selector",
                                "isotype_generating_series",
                                "Simple isomorphism types",
                                structure_class)



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.modules
r"""
Tensor Algebras

AUTHORS:

- Travis Scrimshaw (2014-01-24): Initial version

.. TODO::

    - Coerce to/from free algebra.
"""

#*****************************************************************************
#  Copyright (C) 2014 Travis Scrimshaw <tscrim at ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.categories.algebras import Algebras
from sage.categories.pushout import ConstructionFunctor
from sage.categories.graded_hopf_algebras_with_basis import GradedHopfAlgebrasWithBasis
from sage.categories.homset import Hom
from sage.categories.morphism import Morphism
from sage.categories.modules import Modules
from sage.categories.tensor import tensor
from sage.combinat.free_module import CombinatorialFreeModule, CombinatorialFreeModule_Tensor
from sage.monoids.indexed_free_monoid import IndexedFreeMonoid
from sage.misc.cachefunc import cached_method
from sage.sets.family import Family


class TensorAlgebra(CombinatorialFreeModule):
    r"""
    The tensor algebra `T(M)` of a module `M`.

    Let `\{ b_i \}_{i \in I}` be a basis of the `R`-module `M`. Then the
    tensor algebra `T(M)` of `M` is an associative `R`-algebra, with a
    basis consisting of all tensors of the form
    `b_{i_1} \otimes b_{i_2} \otimes \cdots \otimes b_{i_n}` for
    nonnegative integers `n` and `n`-tuples
    `(i_1, i_2, \ldots, i_n) \in I^n`. The product of `T(M)` is given by

    .. MATH::

        (b_{i_1} \otimes \cdots \otimes b_{i_m}) \cdot (b_{j_1} \otimes
        \cdots \otimes b_{j_n}) = b_{i_1} \otimes \cdots \otimes b_{i_m}
        \otimes b_{j_1} \otimes \cdots \otimes b_{j_n}.

    As an algebra, it is generated by the basis vectors `b_i` of `M`. It
    is an `\NN`-graded `R`-algebra, with the degree of each `b_i` being
    `1`.

    It also has a Hopf algebra structure: The comultiplication is the
    unique algebra morphism `\delta : T(M) \to T(M) \otimes T(M)` defined
    by:

    .. MATH::

        \delta(b_i) = b_i \otimes 1 + 1 \otimes b_i

    (where the `\otimes` symbol here forms tensors in
    `T(M) \otimes T(M)`, not inside `T(M)` itself). The counit is the
    unique algebra morphism `T(M) \to R` sending each `b_i` to `0`. Its
    antipode `S` satisfies

    .. MATH::

        S(b_{i_1} \otimes \cdots \otimes b_{i_m}) = (-1)^m (b_{i_m} \otimes
        \cdots \otimes b_{i_1}).

    This is a connected graded cocommutative Hopf algebra.

    REFERENCES:

    - :wikipedia:`Tensor_algebra`

    .. SEEALSO::

        :class:`TensorAlgebra`

    EXAMPLES::

        sage: C = CombinatorialFreeModule(QQ, ['a','b','c'])
        sage: TA = TensorAlgebra(C)
        sage: TA.dimension()
        +Infinity
        sage: TA.base_ring()
        Rational Field
        sage: TA.algebra_generators()
        Finite family {'a': B['a'], 'b': B['b'], 'c': B['c']}
    """
    def __init__(self, M, prefix='T', category=None, **options):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: C = CombinatorialFreeModule(QQ, ['a','b','c'])
            sage: TA = TensorAlgebra(C)
            sage: TestSuite(TA).run()
            sage: m = SymmetricFunctions(QQ).m()
            sage: Tm = TensorAlgebra(m)
            sage: TestSuite(Tm).run()
        """
        self._base_module = M
        R = M.base_ring()
        category = GradedHopfAlgebrasWithBasis(R.category()).or_subcategory(category)

        CombinatorialFreeModule.__init__(self, R, IndexedFreeMonoid(M.indices()),
                                         prefix=prefix, category=category, **options)

        # the following is not the best option, but it's better than nothing.
        self._print_options['tensor_symbol'] = options.get('tensor_symbol', tensor.symbol)

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: C = CombinatorialFreeModule(QQ, ['a','b','c'])
            sage: TensorAlgebra(C)
            Tensor Algebra of Free module generated by {'a', 'b', 'c'} over Rational Field
        """
        return "Tensor Algebra of {}".format(self._base_module)

    def _repr_term(self, m):
        """
        Return a string of representation of the term indexed by ``m``.

        TESTS::

            sage: C = CombinatorialFreeModule(QQ, ['a','b','c'])
            sage: TA = TensorAlgebra(C)
            sage: s = TA(['a','b','c']).leading_support()
            sage: TA._repr_term(s)
            "B['a'] # B['b'] # B['c']"
            sage: s = TA(['a']*3 + ['b']*2 + ['a','c','b']).leading_support()
            sage: TA._repr_term(s)
            "B['a'] # B['a'] # B['a'] # B['b'] # B['b'] # B['a'] # B['c'] # B['b']"

            sage: I = TA.indices()
            sage: TA._repr_term(I.one())
            '1'
        """
        if len(m) == 0:
            return '1'
        symb = self._print_options['tensor_symbol']
        if symb is None:
            symb = tensor.symbol
        return symb.join(self._base_module._repr_term(k) for k,e in m._monomial for i in range(e))

    def _latex_term(self, m):
        r"""
        Return a latex representation of the term indexed by ``m``.

        TESTS::

            sage: C = CombinatorialFreeModule(QQ, ['a','b','c'])
            sage: TA = TensorAlgebra(C)
            sage: s = TA(['a','b','c']).leading_support()
            sage: TA._latex_term(s)
            'B_{a} \\otimes B_{b} \\otimes B_{c}'

            sage: I = TA.indices()
            sage: TA._latex_term(I.one())
            '1'
        """
        if len(m) == 0:
            return '1'
        symb = " \\otimes "
        return symb.join(self._base_module._latex_term(k) for k,e in m._monomial for i in range(e))

    def _ascii_art_term(self, m):
        """
        Return an ascii art representation of the term indexed by ``m``.

        TESTS::

            sage: C = CombinatorialFreeModule(QQ, Partitions())
            sage: TA = TensorAlgebra(C)
            sage: s = TA([Partition([3,2,2,1]), Partition([3])]).leading_support()
            sage: TA._ascii_art_term(s)
            B    # B
             ***    ***
             **
             **
             *
            sage: s = TA([Partition([3,2,2,1])]*2 + [Partition([3])]*3 + [Partition([1])]*2).leading_support()
            sage: t = TA._ascii_art_term(s); t
            B    # B    # B    # B    # B    # B  # B
             ***    ***    ***    ***    ***    *    *
             **     **
             **     **
             *      *
            sage: t._breakpoints
            [7, 14, 21, 28, 35, 40]

            sage: I = TA.indices()
            sage: TA._ascii_art_term(I.one())
            '1'
        """
        if len(m) == 0:
            return '1'
        from sage.typeset.ascii_art import AsciiArt, ascii_art
        symb = self._print_options['tensor_symbol']
        if symb is None:
            symb = tensor.symbol
        M = self._base_module
        return ascii_art(*(M._ascii_art_term(k)
                           for k, e in m._monomial for _ in range(e)),
                         sep=AsciiArt([symb], breakpoints=[len(symb)]))

    def _element_constructor_(self, x):
        """
        Construct an element of ``self``.

        EXAMPLES::

            sage: C = CombinatorialFreeModule(QQ, ['a','b','c'])
            sage: TA = TensorAlgebra(C)
            sage: TA(['a','b','c'])
            B['a'] # B['b'] # B['c']
            sage: TA(['a','b','b'])
            B['a'] # B['b'] # B['b']
            sage: TA(['a','b','c']) + TA(['a'])
            B['a'] + B['a'] # B['b'] # B['c']
            sage: TA(['a','b','c']) + TA(['a','b','a'])
            B['a'] # B['b'] # B['a'] + B['a'] # B['b'] # B['c']
            sage: TA(['a','b','c']) + TA(['a','b','c'])
            2*B['a'] # B['b'] # B['c']
            sage: TA(C.an_element())
            2*B['a'] + 2*B['b'] + 3*B['c']
        """
        FM = self._indices
        if isinstance(x, (list, tuple)):
            x = FM.prod(FM.gen(elt) for elt in x)
            return self.monomial(x)
        if x in FM._indices:
            return self.monomial(FM.gen(x))
        if x in self._base_module:
            return self.sum_of_terms((FM.gen(k), v) for k,v in x)
        return CombinatorialFreeModule._element_constructor_(self, x)

    def _tensor_constructor_(self, elts):
        """
        Construct an element of ``self`` that is the tensor product of
        the list of base module elements ``elts``.

        TESTS::

            sage: C = CombinatorialFreeModule(ZZ, ['a','b'])
            sage: TA = TensorAlgebra(C)
            sage: x = C.an_element(); x
            2*B['a'] + 2*B['b']
            sage: TA._tensor_constructor_([x, x])
            4*B['a'] # B['a'] + 4*B['a'] # B['b']
             + 4*B['b'] # B['a'] + 4*B['b'] # B['b']
            sage: y = C.monomial('b') + 3*C.monomial('a')
            sage: TA._tensor_constructor_([x, y])
            6*B['a'] # B['a'] + 2*B['a'] # B['b'] + 6*B['b'] # B['a']
             + 2*B['b'] # B['b']
            sage: TA._tensor_constructor_([y]) == y
            True
            sage: TA._tensor_constructor_([x]) == x
            True
            sage: TA._tensor_constructor_([]) == TA.one()
            True
        """
        if not elts:
            return self.one()

        zero = self.base_ring().zero()
        I = self._indices
        cur = {I.gen(k): v for k,v in elts[0]}
        for x in elts[1:]:
            next = {}
            for k,v in cur.items():
                for m,c in x:
                    i = k * I.gen(m)
                    next[i] = cur.get(i, zero) + v * c
            cur = next
        return self._from_dict(cur)

    def _coerce_map_from_(self, R):
        """
        Return ``True`` if there is a coercion from ``R`` into ``self`` and
        ``False`` otherwise.  The things that coerce into ``self`` are:

        - Anything with a coercion into ``self.base_ring()``.

        - Anything with a coercion into the base module of ``self``.

        - A tensor algebra whose base module has a coercion into the base
          module of ``self``.

        - A tensor module whose factors have a coercion into the base
          module of ``self``.

        TESTS::

            sage: C = CombinatorialFreeModule(ZZ, Set([1,2]))
            sage: TAC = TensorAlgebra(C)
            sage: TAC.has_coerce_map_from(ZZ)
            True
            sage: TAC(1) == TAC.one()
            True
            sage: TAC.has_coerce_map_from(C)
            True
            sage: c = C.monomial(2)
            sage: TAC(c)
            B[2]
            sage: d = C.monomial(1)
            sage: TAC(c) * TAC(d)
            B[2] # B[1]
            sage: TAC(c-d) * TAC(c+d)
            -B[1] # B[1] - B[1] # B[2] + B[2] # B[1] + B[2] # B[2]

            sage: TCC = tensor((C,C))
            sage: TAC.has_coerce_map_from(TCC)
            True
            sage: TAC(tensor([c, d]))
            B[2] # B[1]

        ::

            sage: D = CombinatorialFreeModule(ZZ, Set([2,4]))
            sage: TAD = TensorAlgebra(D)
            sage: f = C.module_morphism(on_basis=lambda x: D.monomial(2*x), codomain=D)
            sage: f.register_as_coercion()

            sage: TCD = tensor((C,D))
            sage: TAD.has_coerce_map_from(TCC)
            True
            sage: TAD.has_coerce_map_from(TCD)
            True
            sage: TAC.has_coerce_map_from(TCD)
            False
            sage: TAD.has_coerce_map_from(TAC)
            True
            sage: TAD(3 * TAC([1, 2, 2, 1, 1]))
            3*B[2] # B[4] # B[4] # B[2] # B[2]
        """
        # Base ring coercions
        self_base_ring = self.base_ring()
        if self_base_ring == R:
            return BaseRingLift(Hom(self_base_ring, self))
        if self_base_ring.has_coerce_map_from(R):
            return BaseRingLift(Hom(self_base_ring, self)) * self_base_ring.coerce_map_from(R)

        M = self._base_module
        # Base module coercions
        if R == M:
            return True
        if M.has_coerce_map_from(R):
            phi = M.coerce_map_from(R)
            return self.coerce_map_from(M) * phi

        # Tensor algebra coercions
        if isinstance(R, TensorAlgebra) and M.has_coerce_map_from(R._base_module):
            RM = R._base_module
            phi = M.coerce_map_from(RM)
            return R.module_morphism(lambda m: self._tensor_constructor_(
                                               [phi(RM.monomial(k)) for k in m.to_word_list()]),
                                     codomain=self)

        # Coercions from tensor products
        if (R in Modules(self_base_ring).WithBasis().TensorProducts()
                and isinstance(R, CombinatorialFreeModule_Tensor)
                and all(M.has_coerce_map_from(RM) for RM in R._sets)):
            modules = R._sets
            vector_map = [M.coerce_map_from(RM) for RM in R._sets]
            return R.module_morphism(lambda x: self._tensor_constructor_(
                                               [vector_map[i](M.monomial(x[i]))
                                                for i,M in enumerate(modules)]),
                                     codomain=self)

        return super()._coerce_map_from_(R)

    def construction(self):
        """
        Return the functorial construction of ``self``.

        EXAMPLES::

            sage: C = CombinatorialFreeModule(ZZ, ['a','b','c'])
            sage: TA = TensorAlgebra(C)
            sage: f, M = TA.construction()
            sage: M == C
            True
            sage: f(M) == TA
            True
        """
        return (TensorAlgebraFunctor(self.category().base()), self._base_module)

    def degree_on_basis(self, m):
        """
        Return the degree of the simple tensor ``m``, which is its length
        (thought of as an element in the free monoid).

        EXAMPLES::

            sage: C = CombinatorialFreeModule(QQ, ['a','b','c'])
            sage: TA = TensorAlgebra(C)
            sage: s = TA(['a','b','c']).leading_support(); s
            F['a']*F['b']*F['c']
            sage: TA.degree_on_basis(s)
            3
        """
        return m.length()

    def base_module(self):
        """
        Return the base module of ``self``.

        EXAMPLES::

            sage: C = CombinatorialFreeModule(QQ, ['a','b','c'])
            sage: TA = TensorAlgebra(C)
            sage: TA.base_module() is C
            True
        """
        return self._base_module

    @cached_method
    def one_basis(self):
        r"""
        Return the empty word, which indexes the `1` of this algebra.

        EXAMPLES::

            sage: C = CombinatorialFreeModule(QQ, ['a','b','c'])
            sage: TA = TensorAlgebra(C)
            sage: TA.one_basis()
            1
            sage: TA.one_basis().parent()
            Free monoid indexed by {'a', 'b', 'c'}
            sage: m = SymmetricFunctions(QQ).m()
            sage: Tm = TensorAlgebra(m)
            sage: Tm.one_basis()
            1
            sage: Tm.one_basis().parent()
            Free monoid indexed by Partitions
        """
        return self._indices.one()

    @cached_method
    def algebra_generators(self):
        r"""
        Return the generators of this algebra.

        EXAMPLES::

            sage: C = CombinatorialFreeModule(QQ, ['a','b','c'])
            sage: TA = TensorAlgebra(C)
            sage: TA.algebra_generators()
            Finite family {'a': B['a'], 'b': B['b'], 'c': B['c']}
            sage: m = SymmetricFunctions(QQ).m()
            sage: Tm = TensorAlgebra(m)
            sage: Tm.algebra_generators()
            Lazy family (generator(i))_{i in Partitions}
        """
        return Family(self._indices.indices(),
                      lambda i: self.monomial(self._indices.gen(i)),
                      name='generator')

    gens = algebra_generators

    def product_on_basis(self, a, b):
        r"""
        Return the product of the basis elements indexed by ``a`` and
        ``b``, as per
        :meth:`AlgebrasWithBasis.ParentMethods.product_on_basis()`.

        INPUT:

        - ``a``, ``b`` -- basis indices

        EXAMPLES::

            sage: C = CombinatorialFreeModule(QQ, ['a','b','c'])
            sage: TA = TensorAlgebra(C)
            sage: I = TA.indices()
            sage: g = I.gens()
            sage: TA.product_on_basis(g['a']*g['b'], g['a']*g['c'])
            B['a'] # B['b'] # B['a'] # B['c']
        """
        return self.monomial(a * b)

    def counit(self, x):
        """
        Return the counit of ``x``.

        INPUT:

        - ``x`` -- an element of ``self``

        EXAMPLES::

            sage: C = CombinatorialFreeModule(QQ, ['a','b','c'])
            sage: TA = TensorAlgebra(C)
            sage: x = TA(['a','b','c'])
            sage: TA.counit(x)
            0
            sage: TA.counit(x + 3)
            3
        """
        return x[self.one_basis()]

    def antipode_on_basis(self, m):
        """
        Return the antipode of the simple tensor indexed by ``m``.

        EXAMPLES::

            sage: C = CombinatorialFreeModule(QQ, ['a','b','c'])
            sage: TA = TensorAlgebra(C)
            sage: s = TA(['a','b','c']).leading_support()
            sage: TA.antipode_on_basis(s)
            -B['c'] # B['b'] # B['a']
            sage: t = TA(['a', 'b', 'b', 'b']).leading_support()
            sage: TA.antipode_on_basis(t)
            B['b'] # B['b'] # B['b'] # B['a']
        """
        m = self._indices(reversed(m._monomial))
        R = self.base_ring()
        if len(m) % 2 == 1:
            return self.term(m, -R.one())
        else:
            return self.term(m, R.one())

    def coproduct_on_basis(self, m):
        r"""
        Return the coproduct of the simple tensor indexed by ``m``.

        EXAMPLES::

            sage: C = CombinatorialFreeModule(QQ, ['a','b','c'])
            sage: TA = TensorAlgebra(C, tensor_symbol="(X)")
            sage: TA.coproduct_on_basis(TA.one_basis())
            1 # 1
            sage: I = TA.indices()
            sage: ca = TA.coproduct_on_basis(I.gen('a')); ca
            1 # B['a'] + B['a'] # 1
            sage: s = TA(['a','b','c']).leading_support()
            sage: cp = TA.coproduct_on_basis(s); cp
            1 # B['a'](X)B['b'](X)B['c'] + B['a'] # B['b'](X)B['c']
             + B['a'](X)B['b'] # B['c'] + B['a'](X)B['b'](X)B['c'] # 1
             + B['a'](X)B['c'] # B['b'] + B['b'] # B['a'](X)B['c']
             + B['b'](X)B['c'] # B['a'] + B['c'] # B['a'](X)B['b']

        We check that `\Delta(a \otimes b \otimes c) =
        \Delta(a) \Delta(b) \Delta(c)`::

            sage: cb = TA.coproduct_on_basis(I.gen('b'))
            sage: cc = TA.coproduct_on_basis(I.gen('c'))
            sage: cp == ca * cb * cc
            True
        """
        S = self.tensor_square()
        if len(m) == 0:
            return S.one()

        if len(m) == 1:
            ob = self.one_basis()
            return S.sum_of_monomials([(m, ob), (ob, m)])

        I = self._indices
        m_word = [k for k,e in m._monomial for dummy in range(e)]
        ob = self.one_basis()
        return S.prod(S.sum_of_monomials([(I.gen(x), ob), (ob, I.gen(x))])
                      for x in m_word)

        # TODO: Implement a coproduct using shuffles.
        # This isn't quite right:
        #from sage.combinat.words.word import Word
        #k = len(m)
        #return S.sum_of_monomials( (I.prod(I.gen(m_word[i]) for i in w[:p]),
        #                            I.prod(I.gen(m_word[i]) for i in w[p:]))
        #                          for p in range(k+1)
        #                          for w in Word(range(p)).shuffle(range(p, k)) )

#####################################################################
# TensorAlgebra functor


class TensorAlgebraFunctor(ConstructionFunctor):
    r"""
    The tensor algebra functor.

    Let `R` be a unital ring. Let `V_R` and `A_R` be the categories of
    `R`-modules and `R`-algebras respectively. The functor
    `T : V_R \to A_R` sends an `R`-module `M` to the tensor
    algebra `T(M)`. The functor `T` is left-adjoint to the forgetful
    functor `F : A_R \to V_R`.

    INPUT:

    - ``base`` -- the base `R`
    """
    # We choose a larger (functor) rank than most ConstructionFunctors
    #   since this should be applied after all of the module functors
    rank = 20

    def __init__(self, base):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: from sage.algebras.tensor_algebra import TensorAlgebraFunctor
            sage: F = TensorAlgebraFunctor(Rings())
            sage: TestSuite(F).run()
        """
        ConstructionFunctor.__init__(self, Modules(base), Algebras(base))

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: from sage.algebras.tensor_algebra import TensorAlgebraFunctor
            sage: TensorAlgebraFunctor(Rings())
            Tensor algebra functor on modules over rings
            sage: TensorAlgebraFunctor(QQ)
            Tensor algebra functor on vector spaces over Rational Field
        """
        return "Tensor algebra functor on {}".format(self.domain()._repr_object_names())

    def _apply_functor(self, M):
        """
        Construct the tensor algebra `T(M)`.

        EXAMPLES::

            sage: from sage.algebras.tensor_algebra import TensorAlgebraFunctor
            sage: C = CombinatorialFreeModule(QQ, ['a','b','c'])
            sage: F = TensorAlgebraFunctor(QQ)
            sage: F._apply_functor(C)
            Tensor Algebra of Free module generated by {'a', 'b', 'c'} over Rational Field
        """
        if M not in self.domain().WithBasis():
            raise NotImplementedError("currently only for modules with basis")
        return TensorAlgebra(M)

    def _apply_functor_to_morphism(self, f):
        """
        Apply ``self`` to a morphism ``f`` in the domain of ``self``.

        EXAMPLES::

            sage: from sage.algebras.tensor_algebra import TensorAlgebraFunctor
            sage: C = CombinatorialFreeModule(QQ, ['a','b','c'])
            sage: D = CombinatorialFreeModule(QQ, ['x','y'])
            sage: on_basis = lambda m: C.term('a', 2) + C.monomial('b') if m == 'x' else sum(C.basis())
            sage: phi = D.module_morphism(on_basis, codomain=C); phi
            Generic morphism:
              From: Free module generated by {'x', 'y'} over Rational Field
              To:   Free module generated by {'a', 'b', 'c'} over Rational Field
            sage: list(map(phi, D.basis()))
            [2*B['a'] + B['b'], B['a'] + B['b'] + B['c']]
            sage: F = TensorAlgebraFunctor(QQ)
            sage: Tphi = F._apply_functor_to_morphism(phi); Tphi
            Generic morphism:
              From: Tensor Algebra of Free module generated by {'x', 'y'} over Rational Field
              To:   Tensor Algebra of Free module generated by {'a', 'b', 'c'} over Rational Field
            sage: G = F(D).algebra_generators()
            sage: list(map(Tphi, G))
            [2*B['a'] + B['b'], B['a'] + B['b'] + B['c']]
            sage: Tphi(sum(G))
            3*B['a'] + 2*B['b'] + B['c']
            sage: Tphi(G['x'] * G['y'])
            2*B['a'] # B['a'] + 2*B['a'] # B['b'] + 2*B['a'] # B['c']
             + B['b'] # B['a'] + B['b'] # B['b'] + B['b'] # B['c']
        """
        DB = f.domain()
        D = self(DB)
        C = self(f.codomain())
        phi = lambda m: C._tensor_constructor_([f(DB.monomial(k))
                                                for k in m.to_word_list()])
        return D.module_morphism(phi, codomain=C)

#####################################################################
# Lift map from the base ring


class BaseRingLift(Morphism):
    r"""
    Morphism `R \to T(M)` which identifies the base ring `R` of a tensor
    algebra `T(M)` with the `0`-th graded part of `T(M)`.
    """
    def _call_(self, x):
        """
        Construct the image of ``x``.

        TESTS::

            sage: C = CombinatorialFreeModule(QQ, Set([1,2]))
            sage: TA = TensorAlgebra(C)
            sage: TA(ZZ(2))
            2
        """
        T = self.codomain()
        R = T.base_ring()
        return T.term(T.indices().one(), R(x))
